package info.textgrid.middleware;

import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import javax.xml.datatype.DatatypeConfigurationException;
import org.apache.commons.logging.LogFactory;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.action.search.SearchScrollRequest;
import org.elasticsearch.action.search.SearchType;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.common.Strings;
import org.elasticsearch.common.unit.TimeValue;
import org.elasticsearch.index.query.QueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.json.JSONObject;
import info.textgrid.middleware.common.TextGridMimetypes;
import info.textgrid.middleware.oaipmh.HeaderType;
import info.textgrid.middleware.oaipmh.ListIdentifiersType;
import info.textgrid.middleware.oaipmh.RequestType;
import info.textgrid.middleware.oaipmh.ResumptionTokenType;

/**
 *
 */
public abstract class IdentifierListDelivererAbstract implements IdentifierListDelivererInterface {

  // **
  // FINALS
  // **

  protected static final int lifeTimeResToken = 100;

  // **
  // STATIC
  // **

  private static org.apache.commons.logging.Log log =
      LogFactory.getLog(RecordListDelivererAbstract.class);
  protected static Map<String, Integer> cursorCollector = new Hashtable<String, Integer>();

  // **
  // CLASS
  // **

  protected boolean textgrid;
  protected boolean dariah;
  protected boolean foundItems;
  protected long resultSize;
  protected int searchResponseSize;
  protected ResumptionTokenType resTokenForResponse;
  protected String datestamp;
  protected boolean idExist = true;
  protected OAI_ESClient oaiEsClient;
  protected String[] identifierListFields; // Fields for the elasticsearch request
  protected String rangeField; // Field for the optional range query
  protected String dateOfObjectCreation; // Field for the object creation in the repository
  protected String repositoryObjectURIPrefix; //
  protected String identifierField;
  protected String formatField;
  protected String formatToFilter;

  /**
   * @param textgrid
   * @param dariah
   */
  public IdentifierListDelivererAbstract(boolean textgrid, boolean dariah) {
    this.textgrid = textgrid;
    this.dariah = dariah;
  }

  // **
  // PUBLIC METHODS
  // **

  /**
   * @param request
   * @return
   */
  public static ErrorHandler requestChecker(RequestType request) {

    ErrorHandler result = new ErrorHandler();

    // Check if metadata prefix is existing and valid.
    if (request.getMetadataPrefix() != null
        && !request.getMetadataPrefix().equals(OAIPMHUtilities.OAIDC_PREFIX)
        && !request.getMetadataPrefix().equals(TGConstants.METADATA_IDIOM_PREFIX)
        && !request.getMetadataPrefix().equals(TGConstants.METADATA_OPENAIRE_PREFIX)) {
      result.setError(TGConstants.OAI_METADATA_FORMAT_ERROR,
          "The value of the metadataPrefix: " + request.getMetadataPrefix()
              + " is not supported by the item identified by the value of: "
              + request.getIdentifier());
    }

    // Check if resumptionToken is invalid or existing.
    if (request.getResumptionToken() != null) {
      boolean restokDCExisting = IdentifierListDelivererDC.cursorCollector != null
          && IdentifierListDelivererDC.cursorCollector.containsKey(request.getResumptionToken());
      boolean restokIDIOMExisting = IdentifierListDelivererIDIOM.cursorCollector != null
          && IdentifierListDelivererDC.cursorCollector.containsKey(request.getResumptionToken());
      boolean restok = cursorCollector != null
          && cursorCollector.containsKey(request.getResumptionToken());

      if (!restokDCExisting && !restokIDIOMExisting && !restok) {

        result.setError(TGConstants.OAI_BAD_RESUMPTION_TOKEN, "The value of the "
            + request.getResumptionToken() + " argument is invalid or expired.");
      }
    }

    List<String> errorValues = new ArrayList<String>();

    if (request.getResumptionToken() == null && request.getMetadataPrefix() == null) {
      errorValues.add("metadataPrefix");
    }
    if (request.getIdentifier() != null) {
      errorValues.add("identifier");
    }
    if (errorValues.size() > 0) {
      result.setError(TGConstants.OAI_BAD_ARGUMENT, "The request includes illegal arguments "
          + "or is missing required arguments: " + errorValues);
    }

    return result;
  }

  /**
   * <p>
   * To get the required values for the ListIdentifiers request this function will ask ElasticSearch
   * for a specific textgridUri the values "created" and "textgridUri".
   * </p>
   * 
   * <p>
   * Since the ListIdentifiers request enables the possibility the limit the request in a specific
   * time interval, it is necessary to perform a range query on the ElasticSearch index.</p
   * 
   * @param from : start value for the range query
   * @param to : end value to the range query
   * @return after calling the function "setListIdentifierHeader" the return value is the whole
   *         ListIdentifiers element
   * @throws IOException
   */
  public ListIdentifiersType processIdentifierList(String from, String to, String set,
      String resumptionToken) throws IOException, ParseException {

    // FIXME combine it with with getIdentifierListWithSet and check for empty set

    ListIdentifiersType lit = new ListIdentifiersType();
    QueryBuilder query = setOrNot(set, from, to);
    SearchResponse listListIdentiferValues;

    String[] includes = this.identifierListFields;
    String[] excludes = Strings.EMPTY_ARRAY;
    // TODO: necessary?
    // FetchSourceContext fetchSourceContext = new FetchSourceContext(true, includes, excludes);

    SearchRequest searchRequest =
        new SearchRequest(OAI_ESClient.getEsIndex()).searchType(SearchType.QUERY_THEN_FETCH);

    SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder();
    searchSourceBuilder.query(query);
    searchSourceBuilder.size(this.searchResponseSize);
    searchSourceBuilder.fetchSource(includes, excludes);

    if (resumptionToken != null) {
      SearchScrollRequest scrollRequest = new SearchScrollRequest(resumptionToken);
      scrollRequest.scroll(TimeValue.timeValueSeconds(lifeTimeResToken));
      listListIdentiferValues =
          OAI_ESClient.getEsClient().scroll(scrollRequest, RequestOptions.DEFAULT);
    } else {
      searchRequest.source(searchSourceBuilder);
      searchRequest.scroll(TimeValue.timeValueMinutes(lifeTimeResToken));
      listListIdentiferValues =
          OAI_ESClient.getEsClient().search(searchRequest, RequestOptions.DEFAULT);
    }

    listListIdentiferValues =
        hitHandling(listListIdentiferValues, lit, set, listListIdentiferValues.getScrollId());

    return lit;
  }

  /**
   * @param listFurtherValues
   * @param lit
   * @param set
   * @param resumptionToken
   * @return
   */
  public SearchResponse hitHandling(SearchResponse listFurtherValues, ListIdentifiersType lit,
      String set, String resumptionToken) {

    int i = 0;
    long size = listFurtherValues.getHits().totalHits;
    setResultSize(size);

    log.debug("hits: " + size);

    for (SearchHit hit : listFurtherValues.getHits().getHits()) {
      i++;

      log.debug("hit " + i + ": " + hit.getId());

      // Handle TextGrid.
      if (this.textgrid) {
        String datestamp = hit.getSourceAsMap().get(this.dateOfObjectCreation).toString();
        try {
          datestamp = OAIPMHUtilities.convertDateFormat(datestamp).toXMLFormat();
        } catch (ParseException e) {
          log.error(e.getMessage());
        } catch (DatatypeConfigurationException e) {
          log.error(e.getMessage());
        }
        String identifier = hit.getSourceAsMap().get(this.identifierField).toString();
        lit = setListIdentifierHeader(datestamp, identifier, lit, set);
      }

      // Handle DARIAH.
      else if (this.dariah) {

        log.debug("dateOfObjectCreation: " + this.dateOfObjectCreation);

        JSONObject json = new JSONObject(hit.getSourceAsMap());

        // Get modifiedDate field and convert datestamp.
        this.datestamp = OAIPMHUtilities.fieldLoader(json, this.dateOfObjectCreation);
        try {
          log.debug("datestamp: " + this.datestamp);

          this.datestamp = OAIPMHUtilities.convertDateFormat(this.datestamp).toXMLFormat();

          log.debug("datestamp converted: " + this.datestamp);

        } catch (ParseException e) {
          log.error(e.getMessage());
          // FIXME Go to ERROR state!
        } catch (DatatypeConfigurationException e) {
          log.error(e.getMessage());
          // FIXME Go to ERROR state!
        }

        // Get identifier field.
        String identifier = OAIPMHUtilities.fieldLoader(json, this.identifierField);

        log.debug("identifier: " + identifier);

        lit = setListIdentifierHeader(this.datestamp, identifier, lit, set);
      }
    }

    // Check the need for a resumption token!
    ResumptionTokenType resTokenForResponse = OAIPMHUtilities.getResumptionToken(
        listFurtherValues.getHits().getTotalHits(), resumptionToken, cursorCollector,
        listFurtherValues.getScrollId(), this.searchResponseSize, i);
    if (resTokenForResponse != null) {
      lit.setResumptionToken(resTokenForResponse);
    }

    return listFurtherValues;
  }

  /**
   * @param datestamp
   * @param identifier
   * @param lit
   * @param set
   * @return
   */
  private static ListIdentifiersType setListIdentifierHeader(String datestamp, String identifier,
      ListIdentifiersType lit, String set) {

    // FIXME Delete this function and check for empty set string in setListIdentifierHeaderWithSet

    HeaderType header = new HeaderType();
    header.setDatestamp(datestamp);
    header.setIdentifier(identifier);

    if (set != null) {
      header.getSetSpec().add(set);
    }
    lit.getHeader().add(header);

    return lit;
  }

  /**
   * @param from
   * @param until
   * @param set
   * @param resumptionToken
   * @return
   * @throws IOException
   */
  public List<String> getUriListOfAllEditions(String from, String until, String set,
      String resumptionToken) throws IOException {

    // NOTE Used by OpenAireIdentifierList

    /*
     * InputStream input = null; try { input = new
     * FileInputStream("/etc/dhrep/oaipmh/oaipmh.properties"); } catch (FileNotFoundException e) {
     * // TODO Auto-generated catch block e.printStackTrace(); }
     * 
     * Properties prop = new Properties(); prop.load(input);
     */

    QueryBuilder query;

    log.debug("FROM: " + from);
    log.debug("TO: " + until);

    QueryBuilder rangeQuery = QueryBuilders.rangeQuery("created").from(from).to(until);
    QueryBuilder filterSandBox = QueryBuilders.matchPhraseQuery("nearlyPublished", "true");
    if (set != null) {
      String[] setParts = set.split(":");

      String queryField = "";
      String valueField = "";

      if (setParts[0].equals(TGConstants.SET_FIELD_FOR_TEXTGRID)) {
        queryField = TGConstants.PROJECT_ID_FOR_TEXTGRID;
        valueField = setParts[1];
      }

      log.debug("");

      QueryBuilder matchQuery = QueryBuilders.matchPhraseQuery(queryField, valueField);
      QueryBuilder boolQuery =
          QueryBuilders.boolQuery().must(rangeQuery).must(matchQuery).mustNot(filterSandBox);

      query = boolQuery;
    } else {
      query = QueryBuilders.boolQuery().must(rangeQuery).mustNot(filterSandBox);
    }

    return getContentFromESIndex(query, resumptionToken, set);
  }

  // **
  // PRIVATE METHODS
  // **

  /**
   * @param query
   * @param resumptionToken
   * @param set
   * @return
   * @throws IOException
   */
  private List<String> getContentFromESIndex(QueryBuilder query, String resumptionToken,
      String set) throws IOException {

    List<String> uriList = new ArrayList<String>();
    QueryBuilder recordFilter = QueryBuilders.boolQuery().must(query)
        .must(QueryBuilders.matchPhraseQuery(TGConstants.FORMAT, TextGridMimetypes.EDITION));

    SearchRequest searchRequest = new SearchRequest(OAI_ESClient.getEsIndex());
    SearchSourceBuilder searchSourceBuilder = new SearchSourceBuilder();

    /*
     * InputStream input = null; try { input = new
     * FileInputStream("/etc/dhrep/oaipmh/oaipmh.properties"); } catch (FileNotFoundException e) {
     * // TODO Auto-generated catch block e.printStackTrace(); }
     * 
     * Properties prop = new Properties(); prop.load(input);
     */

    searchSourceBuilder.query(recordFilter);
    searchSourceBuilder.size(100);
    searchRequest.source(searchSourceBuilder);

    SearchResponse scrollResp = new SearchResponse();

    if (resumptionToken != null) {
      SearchScrollRequest scrollRequest = new SearchScrollRequest(resumptionToken);
      scrollRequest.scroll(TimeValue.timeValueHours(24L));

      try {
        scrollResp = OAI_ESClient.getEsClient().scroll(scrollRequest, RequestOptions.DEFAULT);
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    } else {
      searchRequest.source(searchSourceBuilder);
      searchRequest.scroll(TimeValue.timeValueHours(24L));
      try {
        scrollResp = OAI_ESClient.getEsClient().search(searchRequest, RequestOptions.DEFAULT);
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    }

    String scrollID = scrollResp.getScrollId();
    long completeListSize = scrollResp.getHits().totalHits;
    setResultSize(completeListSize);

    if (completeListSize > 0) {

      setFoundItems(true);
      int i = 0;

      for (SearchHit hit : scrollResp.getHits().getHits()) {
        i++;
        if (hit != null && hit.getFields() != null) {
          uriList.add(hit.getSourceAsMap().get(TGConstants.URI).toString());
        }
      }

      if (resumptionToken != null
          && this.resTokenForResponse.getCursor().intValue() >= completeListSize) {
        try {
          cursorCollector.remove(resumptionToken);
        } catch (NullPointerException couldNotRemove) {
          log.info("Could net remove hash value: " + resumptionToken + " from hash map");
        }
        this.resTokenForResponse.setValue("");
      } else {

        this.resTokenForResponse = OAIPMHUtilities.getResumptionToken(completeListSize,
            resumptionToken, cursorCollector, scrollID, this.searchResponseSize, i);
      }

    } else {
      setFoundItems(false);
    }

    return uriList;
  }

  /**
   * @param set
   * @param from
   * @param until
   * @return
   */
  public QueryBuilder setOrNot(String set, String from, String until) {

    QueryBuilder rangeQuery = QueryBuilders.rangeQuery(this.rangeField).from(from).to(until);
    QueryBuilder formatQuery = QueryBuilders.matchPhrasePrefixQuery("format", this.formatToFilter);
    QueryBuilder query;

    if (set != null) {
      String queryField = "";
      String valueField = "";

      if (this.dariah == true) {
        queryField = "descriptiveMetadata.dc:relation";
        valueField = set;
      }

      if (this.textgrid == true) {
        String[] setParts = set.split(":");
        if (setParts[0].equals("project")) {
          queryField = "project.id";
          valueField = setParts[1];
        }
      }

      QueryBuilder matchQuery = QueryBuilders.matchPhraseQuery(queryField, valueField);
      if (this.dariah) {
        query = matchQuery;
      } else {
        query = QueryBuilders.boolQuery().must(matchQuery).must(formatQuery).must(rangeQuery);
      }

    } else {
      if (this.dariah == true)
        query = rangeQuery;
      else
        query = QueryBuilders.boolQuery().must(rangeQuery).must(formatQuery);
    }

    return query;
  }

  // **
  // GETTERS & SETTERS
  // **

  /**
   * @return
   */
  public boolean isFoundItems() {
    return this.foundItems;
  }

  /**
   * @param foundItems
   */
  public void setFoundItems(boolean foundItems) {
    this.foundItems = foundItems;
  }

  /**
   * @return
   */
  public long getResultSize() {
    return this.resultSize;
  }

  /**
   * @param resultSize
   */
  public void setResultSize(long resultSize) {
    this.resultSize = resultSize;
  }

  /**
   * @return
   */
  public ResumptionTokenType getResTokenForResponse() {
    return this.resTokenForResponse;
  }

  /**
   * @param resTokenForResponse
   */
  public void setResTokenForResponse(ResumptionTokenType resTokenForResponse) {
    this.resTokenForResponse = resTokenForResponse;
  }

  /**
   * @return
   */
  public String getFieldForRange() {
    return this.rangeField;
  }

  /**
   * @param fieldForRange
   */
  public void setFieldForRange(String fieldForRange) {
    this.rangeField = fieldForRange;
  }

  /**
   * @return
   */
  public OAI_ESClient getOaiEsClient() {
    return this.oaiEsClient;
  }

  /**
   * @param oaiEsClient
   */
  public void setOaiEsClient(OAI_ESClient oaiEsClient) {
    this.oaiEsClient = oaiEsClient;
  }

  /**
   * @param lit
   * @return
   */
  public ListIdentifiersType getLit(ListIdentifiersType lit) {
    return lit;
  }

  /**
   * @return
   */
  public String getDateOfObjectCreation() {
    return this.dateOfObjectCreation;
  }

  /**
   * @param dateOfObjectCreation
   */
  public void setDateOfObjectCreation(String dateOfObjectCreation) {
    this.dateOfObjectCreation = dateOfObjectCreation;
  }

  /**
   * @param repositoryObjectURIPrefix
   */
  public void setRepositoryObjectURIPrefix(String repositoryObjectURIPrefix) {
    this.repositoryObjectURIPrefix = repositoryObjectURIPrefix;
  }

  /**
   * @return
   */
  public String getRepositoryObjectURIPrefix() {
    return this.repositoryObjectURIPrefix;
  }

  /**
   * @param validItem
   */
  public void setIdExist(boolean validItem) {
    this.idExist = validItem;
  }

  /**
   * @return
   */
  public boolean getIdExist() {
    return this.idExist;
  }

  /**
   * @return
   */
  public String[] getIdentifierListFields() {
    return this.identifierListFields;
  }

  /**
   * @param identifierListFields
   */
  public void setIdentifierListFields(String[] identifierListFields) {
    this.identifierListFields = identifierListFields;
  }

  /**
   * @return
   */
  public String getFormatField() {
    return this.formatField;
  }

  /**
   * @param formatField
   */
  public void setFormatField(String formatField) {
    this.formatField = formatField;
  }

  /**
   * @return
   */
  public String getFormatToFilter() {
    return this.formatToFilter;
  }

  /**
   * @param formatToFilter
   */
  public void setFormatToFilter(String formatToFilter) {
    this.formatToFilter = formatToFilter;
  }

  /**
   * @return
   */
  public String getIdentifierField() {
    return this.identifierField;
  }

  /**
   * @param identifierPrefix
   */
  public void setIdentifierField(String identifierPrefix) {
    this.identifierField = identifierPrefix;
  }

  /**
   * @return
   */
  public int getSearchResponseSize() {
    return this.searchResponseSize;
  }

  /**
   * @param searchResponseSize
   */
  public void setSearchResponseSize(String searchResponseSize) {
    this.searchResponseSize = Integer.parseInt(searchResponseSize);
  }

}
