package info.textgrid.middleware;

import java.io.IOException;
import java.text.ParseException;
import java.util.List;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.datatype.DatatypeConfigurationException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.elasticsearch.common.Strings;
import org.json.JSONObject;
import org.springframework.stereotype.Component;
import info.textgrid.middleware.oaipmh.ContributorType;
import info.textgrid.middleware.oaipmh.DateType;
import info.textgrid.middleware.oaipmh.GetRecordType;
import info.textgrid.middleware.oaipmh.HeaderType;
import info.textgrid.middleware.oaipmh.OAIPMHtype;
import info.textgrid.middleware.oaipmh.ObjectFactory;
import info.textgrid.middleware.oaipmh.OpenaireDescriptionType;
import info.textgrid.middleware.oaipmh.RelatedIdentifierType;
import info.textgrid.middleware.oaipmh.RelationType;
import info.textgrid.middleware.oaipmh.Resource;
import info.textgrid.middleware.oaipmh.Resource.AlternateIdentifiers;
import info.textgrid.middleware.oaipmh.Resource.AlternateIdentifiers.AlternateIdentifier;
import info.textgrid.middleware.oaipmh.Resource.Contributors;
import info.textgrid.middleware.oaipmh.Resource.Contributors.Contributor;
import info.textgrid.middleware.oaipmh.Resource.Creators;
import info.textgrid.middleware.oaipmh.Resource.Creators.Creator;
import info.textgrid.middleware.oaipmh.Resource.Creators.Creator.NameIdentifier;
import info.textgrid.middleware.oaipmh.Resource.Dates;
import info.textgrid.middleware.oaipmh.Resource.Dates.Date;
import info.textgrid.middleware.oaipmh.Resource.Descriptions;
import info.textgrid.middleware.oaipmh.Resource.Descriptions.Description;
import info.textgrid.middleware.oaipmh.Resource.Formats;
import info.textgrid.middleware.oaipmh.Resource.GeoLocations;
import info.textgrid.middleware.oaipmh.Resource.GeoLocations.GeoLocation;
import info.textgrid.middleware.oaipmh.Resource.Identifier;
import info.textgrid.middleware.oaipmh.Resource.RelatedIdentifiers;
import info.textgrid.middleware.oaipmh.Resource.RelatedIdentifiers.RelatedIdentifier;
import info.textgrid.middleware.oaipmh.Resource.ResourceType;
import info.textgrid.middleware.oaipmh.Resource.RightsList;
import info.textgrid.middleware.oaipmh.Resource.RightsList.Rights;
import info.textgrid.middleware.oaipmh.Resource.Sizes;
import info.textgrid.middleware.oaipmh.Resource.Subjects;
import info.textgrid.middleware.oaipmh.Resource.Subjects.Subject;
import info.textgrid.middleware.oaipmh.Resource.Titles;
import info.textgrid.middleware.oaipmh.Resource.Titles.Title;
import info.textgrid.middleware.oaipmh.TitleType;

/**
 * @author Maximilian Brodhun, SUB Göttingen
 * @author Stefan E. Funk, SUB Göttingen
 * @version 2021-07-12
 * @since 2020-06-13
 */
@Component
public class RecordDelivererDATACITE extends RecordDelivererAbstract {

  // Logger for errors and info
  private static Log log = LogFactory.getLog(RecordDelivererDATACITE.class);

  // Container object for the result of the elasticsearch request. This object contains the source
  // of the response
  private JSONObject jsonObj = new JSONObject();

  // This variables contains the field requested in the ElasticSearch Index. The content will be
  // given by a configuration file
  private String oarIdentifierField;
  private String oarSizeField;
  private String[] oarTitleFields;
  private String[] oarDateFields;
  private String[] oarContributorFields;
  private String[] oarCreatorFields;
  private String[] oarLanguageFields;
  private String[] oarAlternateIdentifierFields;
  private String[] oarFormatFields;
  private String[] oarRightsFields;
  private String[] oarDescriptionFields;
  private String[] oarRelatedIdentifierFields;
  private String[] oarWorkFields;
  private String[] oarGeoLocationFields;
  private String[] oarVersionFields;
  private String[] oarSubjectFields;

  private String relationToWorkObject;
  private String handle;

  // This variable contains the XML-content for a whole OpenAire file
  private Resource resource = new Resource();

  /**
   * @param textgrid flag to indicate that the content of the record is context of the TextGrid repo
   * @param dariah flag to indicate that the content of the record is context of the DARIAH repo
   * @throws IOException
   */
  public RecordDelivererDATACITE(boolean textgrid, boolean dariah) throws IOException {
    super(textgrid, dariah);
  }

  /**
   * <p>
   * Building the record XML object for the OAI-PMH response.
   * </p>
   * 
   * @throws ParseException
   * @throws IOException
   */
  public GetRecordType getRecordById(final String idInElasticSearchIndex)
      throws ParseException, DatatypeConfigurationException, IOException {

    log.debug("idInElasticSearchIndex: " + idInElasticSearchIndex);

    // Get ES object.
    this.jsonObj = new JSONObject(OAIPMHUtilities
        .getRcordByIDFromElasticSearch(OAI_ESClient.getEsIndex(),
            idInElasticSearchIndex.replace(this.repositoryObjectURIPrefix, ""), this.fields,
            Strings.EMPTY_ARRAY)
        .getSource());

    System.out.println("jsonObject: " + this.jsonObj);

    // Set response header.
    HeaderType header = new HeaderType();
    header.setDatestamp(OAIPMHUtilities
        .datestampAsString(OAIPMHUtilities.fieldLoader(this.jsonObj, this.dateOfObjectCreation)));

    if (!idInElasticSearchIndex.startsWith(this.repositoryObjectURIPrefix)) {
      header.setIdentifier(this.repositoryObjectURIPrefix + idInElasticSearchIndex);
    } else {
      header.setIdentifier(idInElasticSearchIndex);
    }

    return OAIPMHUtilities.getRecordType(this.setOpenAireRecord(), header);
  }

  /**
   * <p>
   * Put the record from open-aire into the JAXBElement container to add it to an oai-pmh object.
   * </p>
   * 
   * @param idInElasticSearchIndex
   * @return
   * @throws ParseException
   * @throws JAXBException
   * @throws DatatypeConfigurationException
   * @throws IOException
   */
  public JAXBElement<OAIPMHtype> getRecord(String idInElasticSearchIndex)
      throws ParseException, JAXBException, DatatypeConfigurationException, IOException {

    ObjectFactory obf = new ObjectFactory();
    JAXBElement<OAIPMHtype> oaijaxb;
    OAIPMHtype oai = new OAIPMHtype();
    HeaderType header = new HeaderType();

    this.jsonObj =
        new JSONObject(OAIPMHUtilities.getRcordByIDFromElasticSearch(OAI_ESClient.getEsIndex(),
            idInElasticSearchIndex, this.fields, Strings.EMPTY_ARRAY).getSource());

    // Setting the header for the oaipmh response
    header.setDatestamp(OAIPMHUtilities
        .datestampAsString(OAIPMHUtilities.fieldLoader(this.jsonObj, this.dateOfObjectCreation)));

    // The id for the response has to be starts with "textgrid:" or "hdl:". The id in elasticsearch
    // is missing. In consequence is optional in oaipmh request to write textgrid:1233.0 or
    // hdl:blubb-bla or just 1233.0 or blubb-bla. For the response the prefix has to be added.
    if (!idInElasticSearchIndex.startsWith(this.repositoryObjectURIPrefix)) {
      header.setIdentifier(this.repositoryObjectURIPrefix + idInElasticSearchIndex);
    } else {
      header.setIdentifier(idInElasticSearchIndex);
    }

    // It is necessary to add the GetRecordType to the OAIPMH object.
    GetRecordType grt = OAIPMHUtilities.getRecordType(this.setOpenAireRecord(), header);
    oai.setGetRecord(grt);
    oaijaxb = obf.createOAIPMH(oai);
    oaijaxb.setValue(oai);

    return oaijaxb;
  }

  /**
   * Adding the open aire record fields to the openaire record object
   * 
   * @return
   * @throws ParseException
   * @throws DatatypeConfigurationException
   * @throws IOException
   */
  private Resource setOpenAireRecord()
      throws ParseException, DatatypeConfigurationException, IOException {

    this.resource.setTitles(this.addTitles());
    this.resource.setPublisher(addPublisher());
    this.resource.setPublicationYear(this.addPublicationYear());
    this.resource.setDates(this.addDates());
    this.resource.setContributors(this.addContributors());
    this.resource.setLanguage(this.addLanguae());
    this.resource.setAlternateIdentifiers(this.addAlternateIdentifier());
    this.resource.setFormats(this.addFormats());
    this.resource.setRightsList(this.addRights());
    this.resource.setDescriptions(this.addDescriptions());
    this.resource.setRelatedIdentifiers(this.addRelatedIdentifiers());
    this.resource.setIdentifier(this.addIdentifier());
    this.resource.setCreators(this.addCreators());
    this.addGeoLocation();
    this.resource.setResourceType(addResourceType());
    this.resource.setVersion(this.addVersion());

    if (addSubjects().getSubject().get(0).getValue() != null) {
      this.resource.setSubjects(this.addSubjects()); // TODO what if subjects == null?
    }

    this.resource.setSizes(this.addSize());

    return this.resource;
  }

  /**
   * @return
   */
  private Sizes addSize() {
    Sizes sizes = new Sizes();

    log.debug("sizeField: " + this.oarSizeField);

    String size = OAIPMHUtilities.fieldLoader(this.jsonObj, this.oarSizeField) + " Bytes";

    log.debug("size: " + size);

    sizes.getSize().add(size);

    return sizes;
  }

  /**
   * @return
   */
  private Subjects addSubjects() {

    Subjects subjects = new Subjects();

    for (String subjectField : this.oarSubjectFields) {
      Subject subject = new Subject();
      subject.setValue(OAIPMHUtilities.fieldLoader(this.jsonObj, subjectField + ".value"));
      subject.setSchemeURI(OAIPMHUtilities.fieldLoader(this.jsonObj, subjectField + ".id.type"));
      subject
          .setSubjectScheme(OAIPMHUtilities.fieldLoader(this.jsonObj, subjectField + ".id.value"));
      subjects.getSubject().add(subject);
    }
    if (subjects.getSubject().get(0) != null) {
      return subjects;
    } else {
      return null;
    }
  }

  /**
   * @return
   */
  private String addVersion() {

    String version = "";

    for (String versionField : this.oarVersionFields) {
      version = OAIPMHUtilities.fieldLoader(this.jsonObj, versionField);
    }

    return version;
  }

  /**
   * @return
   */
  private static ResourceType addResourceType() {

    ResourceType resourceType = new ResourceType();
    resourceType.setResourceTypeGeneral(info.textgrid.middleware.oaipmh.ResourceType.DATASET);

    // FIXME What about collection? Could editions be seen as collection datasets?

    return resourceType;
  }

  /**
   * 
   */
  private void addGeoLocation() {

    GeoLocations geoLocations = new GeoLocations();

    for (String geoLocationField : this.oarGeoLocationFields) {
      GeoLocation geoLocation = new GeoLocation();
      geoLocation.setGeoLocationPlace(
          OAIPMHUtilities.fieldLoader(this.jsonObj, geoLocationField + ".value"));
      geoLocations.getGeoLocation().add(geoLocation);
    }
    if (geoLocations.getGeoLocation().get(0).getGeoLocationPlace() != null) {
      this.resource.setGeoLocations(geoLocations);
    }

    // return geoLocations;
  }

  /**
   * @return
   * @throws IOException
   */
  private RelatedIdentifiers addRelatedIdentifiers() throws IOException {

    RelatedIdentifiers relatedIdentifiers = new RelatedIdentifiers();

    for (String relatedIdentifierField : this.oarRelatedIdentifierFields) {

      System.out.println("relatedIdentifierField: " + relatedIdentifierField);

      // **
      // TextGrid
      // *

      if (this.textgrid) {
        String[] workFields = {this.handle};
        RelatedIdentifier relatedIdentifier = new RelatedIdentifier();
        relatedIdentifier.setRelatedIdentifierType(RelatedIdentifierType.HANDLE);
        // relatedIdentifier.setRelatedMetadataScheme(TGConstants.TEXTGRID_METADATASCHEME);
        relatedIdentifier.setRelationType(RelationType.IS_PART_OF);
        // relatedIdentifier.setSchemeType("XSD");
        // relatedIdentifier.setSchemeURI(TGConstants.TEXTGRID_METADATASCHEME_URI);

        String uriForWork = OAIPMHUtilities.fieldLoader(this.jsonObj, relatedIdentifierField)
            .replace(this.repositoryObjectURIPrefix, "");

        relatedIdentifier
            .setValue(
                OAIPMHUtilities.fieldLoader(
                    new JSONObject(
                        OAIPMHUtilities.getRcordByIDFromElasticSearch(OAI_ESClient.getEsIndex(),
                            uriForWork, workFields, Strings.EMPTY_ARRAY).getSource()),
                    this.handle));
        relatedIdentifiers.getRelatedIdentifier().add(relatedIdentifier);
      }

      // **
      // DARIAH
      // **

      else if (this.dariah) {
        List<String> relatedIdentiferList =
            OAIPMHUtilities.fieldLoader(this.jsonObj, new String[] {relatedIdentifierField});
        for (String i : relatedIdentiferList) {

          System.out.println("i: " + i);

          RelatedIdentifier relatedID = new RelatedIdentifier();
          relatedID.setValue(i);
          if (i.startsWith("hdl:")) {
            relatedID.setRelatedIdentifierType(RelatedIdentifierType.HANDLE);
          } else if (i.startsWith("doi:")) {
            relatedID.setRelatedIdentifierType(RelatedIdentifierType.DOI);
          } else if (i.startsWith("http")) {
            relatedID.setRelatedIdentifierType(RelatedIdentifierType.URL);
          } else {
            // TODO Check other prefixes and add other values, too?
            // TODO No OTHER value existing! What do we do here, if none of the above apply?
          }
          // TODO We have to check relation type, too? Firstly we take part of a collection
          // here(that's what we put in while publishing).
          relatedID.setRelationType(RelationType.IS_PART_OF);
          relatedIdentifiers.getRelatedIdentifier().add(relatedID);
        }
      }
    }

    return relatedIdentifiers;
  }

  /**
   * @return
   * @throws IOException
   */
  private Descriptions addDescriptions() throws IOException {

    Descriptions descriptions = new Descriptions();

    for (String descriptionField : this.oarDescriptionFields) {

      // **
      // TextGrid
      // **

      if (this.textgrid) {

        Description description = new Description();
        description.setDescriptionType(OpenaireDescriptionType.ABSTRACT);

        log.debug("description: " + descriptionField);
        log.debug("relationToWorkObject: " + this.relationToWorkObject);

        String idForWorkObject =
            OAIPMHUtilities.fieldLoader(this.jsonObj, this.relationToWorkObject);
        if (idForWorkObject != null) {
          idForWorkObject = idForWorkObject.replace(this.repositoryObjectURIPrefix, "");
          JSONObject resultOfFurtherObject = new JSONObject(
              OAIPMHUtilities.getRcordByIDFromElasticSearch(OAI_ESClient.getEsIndex(),
                  idForWorkObject, this.workFields, Strings.EMPTY_ARRAY).getSource());
          String abstractForWork =
              OAIPMHUtilities.fieldLoader(resultOfFurtherObject, descriptionField);

          // TODO Set only language? Is there a textgrid example anywhere?
          description.setLang(abstractForWork);
        }

        descriptions.getDescription().add(description);
      }

      // **
      // DARIAH
      // **

      else if (this.dariah) {
        // Here the version of fieldLoader() is used, that delivers a LIST of strings!
        // TODO Check if you can adapt the other method to only have ONE method left!
        List<String> descriptionsFromMetadata =
            OAIPMHUtilities.fieldLoader(this.jsonObj, new String[] {descriptionField});

        log.debug("descriptions: " + descriptionsFromMetadata);

        // Create abstract for every dc:description.
        for (String d : descriptionsFromMetadata) {
          Description description = new Description();
          description.setDescriptionType(OpenaireDescriptionType.ABSTRACT);
          description.getContent().add(d);
          descriptions.getDescription().add(description);
        }
      }
    }

    return descriptions;
  }

  /**
   * @return
   */
  private RightsList addRights() {

    RightsList rightsList = new RightsList();

    for (String rightsField : this.oarRightsFields) {
      Rights rights = new Rights();
      rights.setRightsURI(OAIPMHUtilities.fieldLoader(this.jsonObj, rightsField + ".licenseUri"));
      rights.setValue(OAIPMHUtilities.fieldLoader(this.jsonObj, rightsField + ".value"));
      rightsList.getRights().add(rights);
    }
    if (rightsList.getRights().get(0).getValue() != null) {
      Rights openAccesRight = new Rights();
      openAccesRight.setRightsURI("info:eu-repo/semantics/openAccess");
      rightsList.getRights().add(openAccesRight);
    }

    return rightsList;
  }

  /**
   * @return
   */
  private Formats addFormats() {

    Formats formats = new Formats();

    for (String formatField : this.oarFormatFields) {
      formats.getFormat().add(OAIPMHUtilities.fieldLoader(this.jsonObj, formatField));
    }

    return formats;
  }

  /**
   * @return
   */
  private AlternateIdentifiers addAlternateIdentifier() {

    AlternateIdentifiers alternateIdentifiers = new AlternateIdentifiers();

    for (String alternateIdentifierField : this.oarAlternateIdentifierFields) {

      // **
      // TextGrid
      // **

      if (this.textgrid) {
        AlternateIdentifier alternateIdentifier = new AlternateIdentifier();
        alternateIdentifier
            .setValue(OAIPMHUtilities.fieldLoader(this.jsonObj, alternateIdentifierField));
        alternateIdentifier.setAlternateIdentifierType(TGConstants.URI_IDENTIIER_TYPE);
        alternateIdentifiers.getAlternateIdentifier().add(alternateIdentifier);
      }

      // **
      // DARIAH
      // **

      else if (this.dariah) {
        List<String> alternateIdentifierList =
            OAIPMHUtilities.fieldLoader(this.jsonObj, new String[] {alternateIdentifierField});
        for (String i : alternateIdentifierList) {
          AlternateIdentifier alternateID = new AlternateIdentifier();
          alternateID.setValue(i);
          if (alternateID.getValue().startsWith("hdl:")) {
            alternateID.setAlternateIdentifierType("HANDLE");
          } else if (alternateID.getValue().startsWith("doi:")) {
            alternateID.setAlternateIdentifierType("DOI");
          } else {
            alternateID.setAlternateIdentifierType("OTHER");
          }
          alternateIdentifiers.getAlternateIdentifier().add(alternateID);
        }
      }
    }

    return alternateIdentifiers;
  }

  /**
   * @return
   */
  private String addLanguae() {

    String language = null;

    // TODO Use only ONE language here? What if we have more then one in DARIAH DC data?

    for (String languageField : this.oarLanguageFields) {
      language = OAIPMHUtilities.fieldLoader(this.jsonObj, languageField);
    }

    return language;
  }

  /**
   * @return
   */
  private Contributors addContributors() {

    Contributors contributors = new Contributors();

    // FIXME fu: Please adapt to DARIAH!

    for (String contributorField : this.oarContributorFields) {
      Contributor contributorInOpenAireRecord = new Contributor();
      if (contributorField.equals("project")) {
        contributorInOpenAireRecord.setContributorName(
            OAIPMHUtilities.fieldLoader(this.jsonObj, contributorField + ".value"));
        contributorInOpenAireRecord.setContributorType(ContributorType.OTHER);
        info.textgrid.middleware.oaipmh.Resource.Contributors.Contributor.NameIdentifier name =
            new info.textgrid.middleware.oaipmh.Resource.Contributors.Contributor.NameIdentifier();
        name.setNameIdentifierScheme("textgrid");
        name.setValue(OAIPMHUtilities.fieldLoader(this.jsonObj, contributorField + ".id"));
        name.setSchemeURI("https://textgridlab.org/schema/textgrid-metadata_2010.xsd");
        contributorInOpenAireRecord.setNameIdentifier(name);
      } else {
        contributorInOpenAireRecord
            .setContributorName(OAIPMHUtilities.fieldLoader(this.jsonObj, contributorField));
        contributorInOpenAireRecord.setContributorType(ContributorType.DATA_MANAGER);
        /*
         * info.textgrid.middleware.oaipmh.Resource.Contributors.Contributor.NameIdentifier name =
         * new info.textgrid.middleware.oaipmh.Resource.Contributors.Contributor.NameIdentifier();
         * name.setNameIdentifierScheme("textgrid"); name.setValue("");
         * name.setSchemeURI("http://www.textgridlab.org/schema/textgrid-metadata_2010.xsd");
         * contributorInOpenAireRecord.setNameIdentifier(name);
         */
      }

      contributors.getContributor().add(contributorInOpenAireRecord);
    }

    return contributors;
  }

  /**
   * @return
   */
  private Dates addDates() {

    Dates dates = new Dates();

    for (String dateField : this.oarDateFields) {

      log.debug("dateField: " + dateField);

      Date dateInOpenAireRecord = new Date();

      // NOTE Date must be taken from oar date fields, not from dateOfObjectCration! That only must
      // be used in the record's header!
      // String date = OAIPMHUtilities.convertDateFormat(OAIPMHUtilities.fieldLoader(this.jsonObj,
      // this.dateOfObjectCreation).toString()).toXMLFormat();
      // dateInOpenAireRecord.setValue(date);

      try {
        // Only one value per field is needed here (I think).
        String dateValue = OAIPMHUtilities
            .convertDateFormat(OAIPMHUtilities.fieldLoader(this.jsonObj, dateField).toString())
            .toXMLFormat();
        dateInOpenAireRecord.setValue(dateValue);

        log.debug("dateValue: " + dateValue);

        // TODO Use extra created, issued, and updated fields in configuration!

        // Filter TextGrid values for date type.
        if (dateField.equals("created")) {
          dateInOpenAireRecord.setDateType(DateType.CREATED);
        }
        if (dateField.equals("issued")) {
          dateInOpenAireRecord.setDateType(DateType.ISSUED);
        }
        if (dateField.equals("lastModified")) {
          dateInOpenAireRecord.setDateType(DateType.UPDATED);
        }

        // Filter DARIAH values for date type.
        if (dateField.equals("administrativeMetadata.dcterms:created")) {
          dateInOpenAireRecord.setDateType(DateType.CREATED);
        }
        if (dateField.equals("administrativeMetadata.dcterms:modified")) {
          dateInOpenAireRecord.setDateType(DateType.UPDATED);
        }
      } catch (ParseException | DatatypeConfigurationException e) {
        log.error("Could not parse date field: " + e.getMessage());
      }

      dates.getDate().add(dateInOpenAireRecord);
    }

    return dates;
  }

  /**
   * @return
   */
  private Identifier addIdentifier() {

    Identifier identifier = new Identifier();
    String idValue = OAIPMHUtilities.fieldLoader(this.jsonObj, this.oarIdentifierField);

    // We must set Handle here for TG, cause we do not have a DOI! So this maybe will be not valid
    // due to Datacite schema validation!
    // We set Handle for DH, too, because we do not have the DOI in administrativeMetadata (yet),
    // seems to be related to the XSD that's creating the ES metadata in esutils!
    // TODO Please check some time!
    identifier.setIdentifierType("Handle");
    identifier.setValue(idValue);

    return identifier;
  }

  /**
   * @return
   */
  private Creators addCreators() {

    Creators creators = new Creators();

    // **
    // TextGrid
    // **

    if (this.textgrid) {

      // TODO: Fields from configFile

      Creator creator = new Creator();
      creator.setCreatorName(OAIPMHUtilities.fieldLoader(this.jsonObj, "edition.agent.value"));
      NameIdentifier nameIdentifier = new NameIdentifier();
      // TODO: creatorID Field from config file
      String creatorID = OAIPMHUtilities.fieldLoader(this.jsonObj,
          "edition.source.bibliographicCitation.author.id");

      if (creatorID != null && creatorID.contains(":")) {
        nameIdentifier.setValue(creatorID.split(":")[1]);
        nameIdentifier.setNameIdentifierScheme(creatorID.split(":")[0]);
        // TODO Check if the scheme URI is correct
        nameIdentifier.setSchemeURI("https://ref.de.dariah.eu/pndsearch/pndquery.xql?id=");
      } else {
        nameIdentifier.setValue(creatorID);
      }
      creator.setNameIdentifier(nameIdentifier);
      nameIdentifier.setSchemeURI("https://ref.de.dariah.eu/pndsearch/pndquery.xql?id=");
      creators.getCreator().add(creator);
    }

    // **
    // DARIAH
    // **

    else if (this.dariah) {

      log.debug("oarCreatorFields: " + this.oarCreatorFields);

      for (String cFields : this.oarCreatorFields) {

        log.debug("creator fields: " + cFields);

        List<String> creatorsList =
            OAIPMHUtilities.fieldLoader(this.jsonObj, this.oarCreatorFields);
        for (String c : creatorsList) {
          Creator creator = new Creator();

          log.debug("creator: " + c);

          creator.setCreatorName(c);
          creators.getCreator().add(creator);
        }
      }

    }

    return creators;
  }

  /**
   * @return
   */
  private Titles addTitles() {

    Titles titles = new Titles();

    log.debug("oarTitleFields: " + this.oarTitleFields);

    for (String titleField : this.oarTitleFields) {

      log.debug("titleField: " + titleField);

      List<String> titleList = OAIPMHUtilities.fieldLoader(this.jsonObj, this.oarTitleFields);
      for (String title : titleList) {
        Title tileInOpenAireRecord = new Title();
        tileInOpenAireRecord.setValue(title);
        tileInOpenAireRecord.setTitleType(TitleType.OTHER);
        titles.getTitle().add(tileInOpenAireRecord);
      }
    }

    return titles;
  }

  /**
   * @return
   */
  private String addPublisher() {

    String result = "";

    /**
     * Publisher fields
     */
    // TODO: fields from config file
    // TODO: for each field
    // FIXME fu: Adapt to DARIAH!
    // return OAIPMHUtilities.fieldLoader(jsonObj,
    // "edition.source.bibliographicCitation.publisher.value");

    if (this.textgrid) {
      result = "TextGrid";
    } else if (this.dariah) {
      result = "DARIAH-DE";
    }

    return result;
  }

  /**
   * @return
   * @throws ParseException
   * @throws DatatypeConfigurationException
   */
  private String addPublicationYear() throws ParseException, DatatypeConfigurationException {
    return Integer.toString(OAIPMHUtilities
        .convertDateFormat(OAIPMHUtilities.fieldLoader(this.jsonObj, this.dateOfObjectCreation))
        .getYear());
  }

  // **
  // GETTERS & SETTERS
  // **

  /**
   * @return
   */
  public String[] getOarTitleFields() {
    return this.oarTitleFields;
  }

  /**
   * @param oarTitleFields
   */
  public void setOarTitleFields(String[] oarTitleFields) {
    this.oarTitleFields = oarTitleFields;
  }

  /**
   * @return
   */
  public String getOarIdentifierField() {
    return this.oarIdentifierField;
  }

  /**
   * @param oarIdentifierField
   */
  public void setOarIdentifierField(String oarIdentifierField) {
    this.oarIdentifierField = oarIdentifierField;
  }

  /**
   * @return
   */
  public Resource getResource() {
    return this.resource;
  }

  /**
   * @param resource
   */
  public void setResource(Resource resource) {
    this.resource = resource;
  }

  /**
   * @return
   */
  public String[] getOarDateFields() {
    return this.oarDateFields;
  }

  /**
   * @param oarDateFields
   */
  public void setOarDateFields(String[] oarDateFields) {
    this.oarDateFields = oarDateFields;
  }

  /**
   * @return
   */
  public String[] getOarCreatorFields() {
    return this.oarCreatorFields;
  }

  /**
   * @param oarCreatorFields
   */
  public void setOarCreatorFields(String[] oarCreatorFields) {
    this.oarCreatorFields = oarCreatorFields;
  }

  /**
   * @return
   */
  public String[] getOarContributorFields() {
    return this.oarContributorFields;
  }

  /**
   * @param oarContributorFields
   */
  public void setOarContributorFields(String[] oarContributorFields) {
    this.oarContributorFields = oarContributorFields;
  }

  /**
   * @return
   */
  public String[] getOarLanguageFields() {
    return this.oarLanguageFields;
  }

  /**
   * @param oarLanguageFields
   */
  public void setOarLanguageFields(String[] oarLanguageFields) {
    this.oarLanguageFields = oarLanguageFields;
  }

  /**
   * @return
   */
  public String[] getOarAlternateIdentifierFields() {
    return this.oarAlternateIdentifierFields;
  }

  /**
   * @param oarAlternateIdentifierFields
   */
  public void setOarAlternateIdentifierFields(String[] oarAlternateIdentifierFields) {
    this.oarAlternateIdentifierFields = oarAlternateIdentifierFields;
  }

  /**
   * @return
   */
  public String[] getOarFormatFields() {
    return this.oarFormatFields;
  }

  /**
   * @param oarFormatFields
   */
  public void setOarFormatFields(String[] oarFormatFields) {
    this.oarFormatFields = oarFormatFields;
  }

  /**
   * @return
   */
  public String[] getOarRightsFields() {
    return this.oarRightsFields;
  }

  /**
   * @param oarRightsFields
   */
  public void setOarRightsFields(String[] oarRightsFields) {
    this.oarRightsFields = oarRightsFields;
  }

  /**
   * @return
   */
  public String[] getOarDescriptionFields() {
    return this.oarDescriptionFields;
  }

  /**
   * @param oarDescriptionFields
   */
  public void setOarDescriptionFields(String[] oarDescriptionFields) {
    this.oarDescriptionFields = oarDescriptionFields;
  }

  /**
   * @return
   */
  public String[] getOarRelatedIdentifierFields() {
    return this.oarRelatedIdentifierFields;
  }

  /**
   * @param oarRelatedIdentifierFields
   */
  public void setOarRelatedIdentifierFields(String[] oarRelatedIdentifierFields) {
    this.oarRelatedIdentifierFields = oarRelatedIdentifierFields;
  }

  /**
   * @return
   */
  public String getRelationToWorkObject() {
    return this.relationToWorkObject;
  }

  /**
   * @param relationToWorkObject
   */
  public void setRelationToWorkObject(String relationToWorkObject) {
    this.relationToWorkObject = relationToWorkObject;
  }

  /**
   * @return
   */
  public String[] getOarWorkFields() {
    return this.oarWorkFields;
  }

  /**
   * @param oarWorkFields
   */
  public void setOarWorkFields(String[] oarWorkFields) {
    this.oarWorkFields = oarWorkFields;
  }

  /**
   * @return
   */
  public String[] getOarGeoLocationFields() {
    return this.oarGeoLocationFields;
  }

  /**
   * @param oarGeoLocationFields
   */
  public void setOarGeoLocationFields(String[] oarGeoLocationFields) {
    this.oarGeoLocationFields = oarGeoLocationFields;
  }

  /**
   * @return
   */
  public String getHandle() {
    return this.handle;
  }

  /**
   * @param handle
   */
  public void setHandle(String handle) {
    this.handle = handle;
  }

  /**
   * @return
   */
  public String[] getOarVersionFields() {
    return this.oarVersionFields;
  }

  /**
   * @param oarVersionFields
   */
  public void setOarVersionFields(String[] oarVersionFields) {
    this.oarVersionFields = oarVersionFields;
  }

  /**
   * @return
   */
  public String[] getOarSubjectFields() {
    return this.oarSubjectFields;
  }

  /**
   * @param oarSubjectFields
   */
  public void setOarSubjectFields(String[] oarSubjectFields) {
    this.oarSubjectFields = oarSubjectFields;
  }

  /**
   * @return
   */
  public String getOarSizeField() {
    return this.oarSizeField;
  }

  /**
   * @param oarSize
   */
  public void setOarSizeField(String oarSize) {
    this.oarSizeField = oarSize;
  }

}
